// Copyright (c) 2015 Uber Technologies, Inc.

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Portions of the code below originally from:
// https://github.com/mapbox/mapbox-gl-js/blob/master/js/ui/handler/scroll_zoom.js
import {PropTypes, Component, createElement} from 'react';
import shallowCompare from 'react-addons-shallow-compare';
import autobind from '../utils/autobind';
import document from 'global/document';
import window from 'global/window';

function noop() {}

var ua = typeof window.navigator !== 'undefined' ?
  window.navigator.userAgent.toLowerCase() : '';
var firefox = ua.indexOf('firefox') !== -1;

// Extract a position from a mouse event
function getMousePosition(el, event) {
  var rect = el.getBoundingClientRect();
  event = event.touches ? event.touches[0] : event;
  return [
    event.clientX - rect.left - el.clientLeft,
    event.clientY - rect.top - el.clientTop
  ];
}

// Extract an array of touch positions from a touch event
function getTouchPositions(el, event) {
  var points = [];
  var rect = el.getBoundingClientRect();
  var touches = getTouches(event);
  for (var i = 0; i < touches.length; i++) {
    points.push([
      touches[i].clientX - rect.left - el.clientLeft,
      touches[i].clientY - rect.top - el.clientTop
    ]);
  }
  return points;
}

// Get relevant touches from event depending on event type (for `touchend` and
// `touchcancel` the property `changedTouches` contains the relevant coordinates)
function getTouches(event) {
  var type = event.type;
  if (type === 'touchend' || type === 'touchcancel') {
    return event.changedTouches;
  }
  return event.touches;
}

// Return the centroid of an array of points
function centroid(positions) {
  var sum = positions.reduce(
    function (acc, elt) { return [acc[0] + elt[0], acc[1] + elt[1]]; },
    [0, 0]
  );
  return [sum[0] / positions.length, sum[1] / positions.length];
}

var propTypes = {
  width: PropTypes.number.isRequired,
  height: PropTypes.number.isRequired,
  onMouseDown: PropTypes.func,
  onMouseDrag: PropTypes.func,
  onMouseRotate: PropTypes.func,
  onMouseUp: PropTypes.func,
  onMouseMove: PropTypes.func,
  onMouseClick: PropTypes.func,
  onTouchStart: PropTypes.func,
  onTouchDrag: PropTypes.func,
  onTouchRotate: PropTypes.func,
  onTouchEnd: PropTypes.func,
  onTouchTap: PropTypes.func,
  onZoom: PropTypes.func,
  onZoomEnd: PropTypes.func
};

var defaultProps = {
  onMouseDown: noop,
  onMouseDrag: noop,
  onMouseRotate: noop,
  onMouseUp: noop,
  onMouseMove: noop,
  onMouseClick: noop,
  onTouchStart: noop,
  onTouchDrag: noop,
  onTouchRotate: noop,
  onTouchEnd: noop,
  onTouchTap: noop,
  onZoom: noop,
  onZoomEnd: noop
};

var Interactions = (function (Component) {
  function Interactions(props) {
    Component.call(this, props);
    this.state = {
      didDrag: false,
      isFunctionKeyPressed: false,
      startPos: null,
      pos: null,
      mouseWheelPos: null
    };
    autobind(this);
  }

  if ( Component ) Interactions.__proto__ = Component;
  Interactions.prototype = Object.create( Component && Component.prototype );
  Interactions.prototype.constructor = Interactions;

  Interactions.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
    return shallowCompare(this, nextProps, nextState);
  };

  Interactions.prototype._getMousePos = function _getMousePos (event) {
    var el = this.refs.container;
    return getMousePosition(el, event);
  };

  Interactions.prototype._getTouchPos = function _getTouchPos (event) {
    var el = this.refs.container;
    var positions = getTouchPositions(el, event);
    return centroid(positions);
  };

  Interactions.prototype._isFunctionKeyPressed = function _isFunctionKeyPressed (event) {
    return Boolean(event.metaKey || event.altKey ||
      event.ctrlKey || event.shiftKey);
  };

  Interactions.prototype._onMouseDown = function _onMouseDown (event) {
    var pos = this._getMousePos(event);
    this.setState({
      didDrag: false,
      startPos: pos,
      pos: pos,
      isFunctionKeyPressed: this._isFunctionKeyPressed(event)
    });
    this.props.onMouseDown({pos: pos});
    document.addEventListener('mousemove', this._onMouseDrag, false);
    document.addEventListener('mouseup', this._onMouseUp, false);
  };

  Interactions.prototype._onTouchStart = function _onTouchStart (event) {
    var pos = this._getTouchPos(event);
    this.setState({
      didDrag: false,
      startPos: pos,
      pos: pos,
      isFunctionKeyPressed: this._isFunctionKeyPressed(event)
    });
    this.props.onTouchStart({pos: pos});
    document.addEventListener('touchmove', this._onTouchDrag, false);
    document.addEventListener('touchend', this._onTouchEnd, false);
  };

  Interactions.prototype._onMouseDrag = function _onMouseDrag (event) {
    var pos = this._getMousePos(event);
    this.setState({pos: pos, didDrag: true});
    if (this.state.isFunctionKeyPressed) {
      var ref = this.state;
      var startPos = ref.startPos;
      this.props.onMouseRotate({pos: pos, startPos: startPos});
    } else {
      this.props.onMouseDrag({pos: pos});
    }
  };

  Interactions.prototype._onTouchDrag = function _onTouchDrag (event) {
    var pos = this._getTouchPos(event);
    this.setState({pos: pos, didDrag: true});
    if (this.state.isFunctionKeyPressed) {
      var ref = this.state;
      var startPos = ref.startPos;
      this.props.onTouchRotate({pos: pos, startPos: startPos});
    } else {
      this.props.onTouchDrag({pos: pos});
    }
    event.preventDefault();
  };

  Interactions.prototype._onMouseUp = function _onMouseUp (event) {
    document.removeEventListener('mousemove', this._onMouseDrag, false);
    document.removeEventListener('mouseup', this._onMouseUp, false);
    var pos = this._getMousePos(event);
    this.setState({pos: pos});
    this.props.onMouseUp({pos: pos});
    if (!this.state.didDrag) {
      this.props.onMouseClick({pos: pos});
    }
  };

  Interactions.prototype._onTouchEnd = function _onTouchEnd (event) {
    document.removeEventListener('touchmove', this._onTouchDrag, false);
    document.removeEventListener('touchend', this._onTouchEnd, false);
    var pos = this._getTouchPos(event);
    this.setState({pos: pos});
    this.props.onTouchEnd({pos: pos});
    if (!this.state.didDrag) {
      this.props.onTouchTap({pos: pos});
    }
  };

  Interactions.prototype._onMouseMove = function _onMouseMove (event) {
    var pos = this._getMousePos(event);
    this.props.onMouseMove({pos: pos});
  };

  /* eslint-disable complexity, max-statements */
  Interactions.prototype._onWheel = function _onWheel (event) {
    event.preventDefault();
    var value = event.deltaY;
    // Firefox doubles the values on retina screens...
    if (firefox && event.deltaMode === window.WheelEvent.DOM_DELTA_PIXEL) {
      value /= window.devicePixelRatio;
    }
    if (event.deltaMode === window.WheelEvent.DOM_DELTA_LINE) {
      value *= 40;
    }

    var type = this.state.mouseWheelType;
    var timeout = this.state.mouseWheelTimeout;
    var lastValue = this.state.mouseWheelLastValue;
    var time = this.state.mouseWheelTime;

    var now = (window.performance || Date).now();
    var timeDelta = now - (time || 0);

    var pos = this._getMousePos(event);
    time = now;

    if (value !== 0 && value % 4.000244140625 === 0) {
      // This one is definitely a mouse wheel event.
      type = 'wheel';
      // Normalize this value to match trackpad.
      value = Math.floor(value / 4);
    } else if (value !== 0 && Math.abs(value) < 4) {
      // This one is definitely a trackpad event because it is so small.
      type = 'trackpad';
    } else if (timeDelta > 400) {
      // This is likely a new scroll action.
      type = null;
      lastValue = value;

      // Start a timeout in case this was a singular event, and delay it by up
      // to 40ms.
      timeout = window.setTimeout(function setTimeout() {
        var _type = 'wheel';
        this._zoom(-this.state.mouseWheelLastValue, this.state.mouseWheelPos);
        this.setState({mouseWheelType: _type});
      }.bind(this), 40);
    } else if (!this._type) {
      // This is a repeating event, but we don't know the type of event just
      // yet.
      // If the delta per time is small, we assume it's a fast trackpad;
      // otherwise we switch into wheel mode.
      type = Math.abs(timeDelta * value) < 200 ? 'trackpad' : 'wheel';

      // Make sure our delayed event isn't fired again, because we accumulate
      // the previous event (which was less than 40ms ago) into this event.
      if (timeout) {
        window.clearTimeout(timeout);
        timeout = null;
        value += lastValue;
      }
    }

    // Slow down zoom if shift key is held for more precise zooming
    if (event.shiftKey && value) {
      value = value / 4;
    }

    // Only fire the callback if we actually know what type of scrolling device
    // the user uses.
    if (type) {
      this._zoom(-value, pos);
    }

    this.setState({
      mouseWheelTime: time,
      mouseWheelPos: pos,
      mouseWheelType: type,
      mouseWheelTimeout: timeout,
      mouseWheelLastValue: lastValue
    });
  };
  /* eslint-enable complexity, max-statements */

  Interactions.prototype._zoom = function _zoom (delta, pos) {
    // Scale by sigmoid of scroll wheel delta.
    var scale = 2 / (1 + Math.exp(-Math.abs(delta / 100)));
    if (delta < 0 && scale !== 0) {
      scale = 1 / scale;
    }
    this.props.onZoom({pos: pos, scale: scale});
    window.clearTimeout(this._zoomEndTimeout);
    this._zoomEndTimeout = window.setTimeout(function _setTimeout() {
      this.props.onZoomEnd();
    }.bind(this), 200);
  };

  Interactions.prototype.render = function render () {
    var ref = this.props;
    var width = ref.width;
    var height = ref.height;

    return createElement.apply(void 0, [ 'div', {
      ref: 'container',
      onMouseMove: this._onMouseMove,
      onMouseDown: this._onMouseDown,
      onTouchStart: this._onTouchStart,
      onContextMenu: this._onMouseDown,
      onWheel: this._onWheel,
      style: {width: width, height: height, position: 'relative'}
    } ].concat( this.props.children ));
  };

  return Interactions;
}(Component));

export default Interactions;

Interactions.displayName = 'Interactions';
Interactions.propTypes = propTypes;
Interactions.defaultProps = defaultProps;
